#ifndef _COCOON_H
#define _COCOON_H

#include <string>
#include <stdexcept>
#include <fstream>
#include <sstream>
#include <complex>
#include <map>
#include <set>

#include "krak/krak.h"
#include "interval/doubleInterval.h"
#include "vectalg/vector.h"
#include "vectalg/matrix.h"
#include "vectalg/norm.h"
#include "map/map.h"
#include "map/c2map.h"
#include "dynset/rect2set.h"
#include "dynset/c1rect2.h"
#include "dynset/c2rect2.h"
#include "covrel/graphhset3D.h"
#include "covrel/graphtset.h"
#include "poincare/poincare.h"
#include "dynsys/c2taylor.h"

typedef std::complex<double> Complex;
typedef capd::covrel::GraphicsHSet3D HSet3D;
typedef capd::covrel::GraphicsTripleSet HSet;

// definitions for dimension 3
typedef Vector<interval,3> IVector3D;
typedef Matrix<interval,3,3> IMatrix3D;
typedef EuclLNorm<IVector3D,IMatrix3D> ENorm3D;
typedef capd::dynset::Rect2Set<IMatrix3D> Rect2Set3D;
typedef capd::dynset::C1rect2<IMatrix3D> C1Rect23D;
typedef capd::dynset::C2rect2<IMatrix3D> C2Rect23D;
typedef capd::map::C2Map<IMatrix3D> Map3D;
typedef capd::dynsys::C2Taylor<Map3D> C2Taylor3D;
typedef capd::poincare::PoincareMap<C2Taylor3D> C2PoincareMap3D;
typedef capd::covrel::GridSet<IMatrix3D> GridSet3D;

// definitions for dimension 4
typedef Vector<interval,4> IVector4D;
typedef Matrix<interval,4,4> IMatrix4D;
typedef EuclLNorm<IVector4D,IMatrix4D> ENorm4D;
typedef capd::dynset::Rect2Set<IMatrix4D> Rect2Set4D;
typedef capd::dynset::C1rect2<IMatrix4D> C1Rect24D;
typedef capd::dynset::C2rect2<IMatrix4D> C2Rect24D;
typedef capd::map::C2Map<IMatrix4D> Map4D;
typedef capd::dynsys::C2Taylor<Map4D> C2Taylor4D; 
typedef capd::poincare::PoincareMap<C2Taylor4D> C2PoincareMap4D;
typedef capd::covrel::GridSet<IMatrix4D> GridSet4D;

struct C2Data{
   Vector<interval,2> iv;
   Matrix<interval,2,2> DP, D2P[2];
};

typedef std::pair<long,long> intPair;

struct LinkC2Data : public C2Data{
   Vector<interval,2> x;
   std::set<intPair> link;
};

typedef std::map<intPair,C2Data> C2DataSet;
typedef std::vector<LinkC2Data> C2DataVector;
typedef bool (*FuncPointer)(const IVector3D&, const IVector3D&);

extern IVector3D Is1, Is2, Iu;
extern const interval IyBif, IcBif;
extern HSet3D N;
extern HSet H0,H1,H2,H3,H4,H5;
extern double maxDelta;

// in file utils.cpp
void findEigenvectors(double Cparam);
void gridSet(std::vector<Rect2Set3D>& GS, IVector3D iv, int gridX, int gridY);
void removeUnnecesaryBlocks(GridSet3D& GS);
void printVector(Vector<interval,2> v);
void coordSystem(Frame& f, double sx, double sy, double x,double y);
void coordSystem3D(Frame *f, double s);
void drawTrapezoid(Frame& f, int color);
void drawVector(Frame &f, const IVector3D& T, int color,int d1=0,int d2=1);

// in file algorithms.cpp
void IHullMatrix( const Matrix<interval,2,2>& DP, Matrix<interval,2,2> D2P[], 
                  Matrix<interval,2,2>& maxDP, Matrix<interval,2,2> maxD2P[]
                );
void IHullMatrix(const IMatrix3D& DP, IMatrix3D D2P[],IMatrix3D& maxDP, IMatrix3D maxD2P[]);
void computeCompositionOfPoincareMap(
      const Matrix<interval,2,2>& DP1,
      const Matrix<interval,2,2>& DP2,
      Matrix<interval,2,2> D2P1[],
      Matrix<interval,2,2> D2P2[],
      Matrix<interval,2,2>& DQ,
      Matrix<interval,2,2> D2Q[]
   );
Matrix<interval,2,2> computeCompositionOfPoincareMap(
      const Matrix<interval,2,2>& DP1,
      const Matrix<interval,2,2>& DP2,
      Matrix<interval,2,2> D2P1[],
      Matrix<interval,2,2> D2P2[]
   );
void saveHalfPoincare(C2PoincareMap3D& PM, IVector3D X, C2Data& c);
void saveHalfPoincare(C2PoincareMap3D& PM, IVector3D X, IMatrix3D& C, IVector3D& r, C2Data& c);
void computeC2ImageOfBox(C2PoincareMap3D& PM, LinkC2Data& it, C2DataSet& halfPoincareData, double meshSize);
void computeImage(C2PoincareMap3D& PM, GridSet3D& GS, FuncPointer fp, Frame& f, int iter, int color);
void computeTimeMap(C2Taylor3D& DS, GridSet3D& GS, FuncPointer fp, Frame* f, int iter, int color);
IMatrix3D computeTimeMap(C2Taylor3D& DS, GridSet3D& GS, const IMatrix3D& M, const IMatrix3D& N,int iter);

// in file conditions.cpp
bool outsideTrapezoid(const IVector3D& iv);
bool outsideTrapezoid(const Vector<interval,2>& Z);
bool outsideH1H2Trapezoid(const IVector3D& iv, const IVector3D&);
bool empty(const IVector3D&, const IVector3D&);
bool monotonicity(const IVector3D& image, const IVector3D& X);
bool leftH3(const IVector3D& iv, const IVector3D&);
bool rightH3(const IVector3D& iv, const IVector3D&);
bool acrossH3(const IVector3D& iv, const IVector3D&);
bool leftH4(const IVector3D& iv, const IVector3D&);
bool rightH4(const IVector3D& iv, const IVector3D&);
bool acrossH4(const IVector3D& iv, const IVector3D&);
bool leftH5(const IVector3D& iv, const IVector3D&);
bool rightH5(const IVector3D& iv, const IVector3D&);
bool acrossH5(const IVector3D& iv, const IVector3D&);
bool leftN(const IVector3D& iv, const IVector3D&);
bool rightN(const IVector3D& iv, const IVector3D&);
bool acrossN(const IVector3D& iv, const IVector3D&);




// --------------------------------------------------
// template functions


// the following function computes Poincare Map and returns their parial derivatives
// up to order 2 in the variables DP and D2P
// the image of point X is stored in X
template<typename PMap>
void computePoincare(PMap& PM, typename PMap::VectorType& X, typename PMap::MatrixType& DP, typename PMap::MatrixType* D2P, int iter=1)
{
   typedef typename PMap::VectorType IVector;
   typedef typename PMap::MatrixType IMatrix;
	IMatrix derivativeOfFlow;
	IMatrix *hessianOfFlow = new IMatrix[IVector::dimension()];

	capd::dynset::C2rect2<IMatrix> theSet(X,EuclLNorm<IVector,IMatrix>());
   for(int i=1;i<=iter;++i)
   {
      if(i==iter)
         PM.setFactor(0.1);
      else
         PM.setFactor(1);
      X = PM(theSet,derivativeOfFlow,hessianOfFlow);
   }
	IVector dT = PM.computeDT(X,derivativeOfFlow);
	DP = PM.computeDP(X,derivativeOfFlow,dT);
	IMatrix d2T = PM.computeD2T(X,derivativeOfFlow,dT,DP,hessianOfFlow);
	PM.computeD2P(X,derivativeOfFlow,dT,DP,hessianOfFlow,d2T,D2P);
   delete [] hessianOfFlow;
}

// the following function computes Poincare Map and returns their parial derivatives
// up to order 2 in the variables DP and D2P
// the image of point X is stored in X
template<typename PMap>
void computePoincare(
      PMap& PM, 
      typename PMap::VectorType& X, 
      typename PMap::MatrixType& C, 
      typename PMap::VectorType& r, 
      typename PMap::MatrixType& DP, 
      typename PMap::MatrixType* D2P, 
      int iter=1
   )
{
   typedef typename PMap::VectorType IVector;
   typedef typename PMap::MatrixType IMatrix;
	IMatrix derivativeOfFlow;
	IMatrix *hessianOfFlow = new IMatrix[IVector::dimension()];

	capd::dynset::C2rect2<IMatrix> theSet(X,C,r,EuclLNorm<IVector,IMatrix>());
   for(int i=1;i<=iter;++i)
   {
      if(i==iter)
         PM.setFactor(0.1);
      else
         PM.setFactor(1);
      X = PM(theSet,derivativeOfFlow,hessianOfFlow);
   }

	IVector dT = PM.computeDT(X,derivativeOfFlow);
	DP = PM.computeDP(X,derivativeOfFlow,dT);
	IMatrix d2T = PM.computeD2T(X,derivativeOfFlow,dT,DP,hessianOfFlow);
	PM.computeD2P(X,derivativeOfFlow,dT,DP,hessianOfFlow,d2T,D2P);
   delete [] hessianOfFlow;
}

// the following function computes Poincare Map and returns their parial derivatives
// up to order 2 in the variables DP and D2P
// the image of point X is stored in X
template<typename PMap>
void computePoincare(
      PMap& PM, 
      typename PMap::VectorType& X, 
      typename PMap::MatrixType& C, 
      typename PMap::VectorType& r, 
      typename PMap::MatrixType& DP, 
      int iter=1
   )
{
   typedef typename PMap::VectorType IVector;
   typedef typename PMap::MatrixType IMatrix;
	IMatrix derivativeOfFlow;

	capd::dynset::C1rect2<IMatrix> theSet(X,C,r,EuclLNorm<IVector,IMatrix>());
   for(int i=1;i<=iter;++i)
   {
      if(i==iter)
         PM.setFactor(0.1);
      else
         PM.setFactor(1);
      X = PM(theSet,derivativeOfFlow);
   }

	IVector dT = PM.computeDT(X,derivativeOfFlow);
	DP = PM.computeDP(X,derivativeOfFlow,dT);
}

// ------------------------------------------------------------------------- //

template<typename ScalarType, int dim>
Matrix<ScalarType,dim,dim> IHull(const Matrix<ScalarType,dim,dim>& a, const Matrix<ScalarType,dim,dim>& b)
{
   Matrix<ScalarType,dim,dim> result=a;
   for(int i=0;i<dim;++i)
      for(int j=0;j<dim;++j)
         result[i][j] = intervalHull(result[i][j],b[i][j]);
   return result;
}

// ------------------------------------------------------------------------- //

// this function verifies if a qudratic form m is positively definite, we assume dim(m)<=3

template<typename MatrixType>
void verifyQuadraticForm(const MatrixType& m)
{
   rootFrame << "The matrix (DFc)^T*Q_1*(DFc)-Q_2 =  " << m;
   int dim = m.numberOfRows();
   interval det = m(1,1);
   rootFrame << "det1 = " << det << "\n";
   if(!(det>0))
      throw std::runtime_error("cone condition not satisfied");
   if(dim>=2)
   {
      det = m(1,1)*m(2,2)-m(1,2)*m(2,1);
      rootFrame << "det2 = " << det << "\n";
      if(!(det>0))
         throw std::runtime_error("cone condition not satisfied");
   }
   if(dim>=3)
   {
      det = m(1,1)*m(2,2)*m(3,3)
          + m(1,2)*m(2,3)*m(3,1)
          + m(1,3)*m(2,1)*m(3,2)
          - m(1,3)*m(2,2)*m(3,1)
          - m(1,1)*m(2,3)*m(3,2)
          - m(1,2)*m(2,1)*m(3,3);
      rootFrame << "det3 = " << det << "\n";
      if(!(det>0))
         throw std::runtime_error("cone condition not satisfied");
   }
   rootFrame << "is positively definite\n\n";
}
#endif
