#include "cocoon.h"

// the following function computes eigenvectors at the equilibrium
// by explicit formula. The computations are performed in floating point 
// arithmetics

void findEigenvectors(double Cparam)
{
	double mian1 = 27*sqrt(2.)*Cparam + sqrt(108+1458*Cparam*Cparam);
	mian1 = exp(log(mian1)/3.);

	double mian2 = -6*exp(log(2.)/3) + exp(log(4.)/3)*mian1*mian1;

	Complex temp = -Complex(1.,sqrt(3.))/(exp(log(4.)/3)*mian1)
                 + Complex(1.,-sqrt(3.))*mian1/(6*exp(log(2.)/3.));

	Complex A = -1./(sqrt(2.)*Cparam*temp) - temp/(sqrt(2.)*Cparam);
	Complex B = -1./temp;
	Complex C = 1.;

	Is1[0] = real(A);
	Is2[0] = imag(A);
	Is1[1] = real(B);
	Is2[1] = imag(B);
	Is1[2] = real(C);
	Is2[2] = imag(C);

	Iu[0] = 3.*sqrt(2.)*mian1/(Cparam*mian2) + mian2/(6*sqrt(2.)*Cparam*mian1);
	Iu[1] = 6.*mian1/mian2;
	Iu[2] = 1.;
   }


// ------------------------------------------------------------------------- //

void gridSet(std::vector<Rect2Set3D>& GS, IVector3D iv, int gridX, int gridY)
{
   iv[2] = interval(0);
   IVector3D left = leftVector(iv), right = rightVector(iv);
   IVector3D skipX((right[0]-left[0])/gridX,interval(0),interval(0));
   IVector3D skipY(interval(0),(right[1]-left[1])/gridY,interval(0));
   IVector3D skip = (skipX+skipY)*interval(0,1);
   
   GS.resize(gridX*gridY,Rect2Set3D(3));
   for(int i=0;i<gridX;++i)
      for(int j=0;j<gridY;j++)
      {
         IVector3D X = left + i*skipX + j*skipY + skip;
         GS[i*gridY+j] = Rect2Set3D(midVector(X),X-midVector(X));
      }
}

// ------------------------------------------------------------------------- //

void removeUnnecesaryBlocks(GridSet3D& GS)
{
   std::vector<GridSet3D::VectorType> newCenters;
   GridSet3D::VectorType r = GS.C * GS.r;
   GridSet3D::iterator b=GS.begin(), e=GS.end();
   while(b!=e)
   {
      GridSet3D::VectorType iv = (*b) + r;
      if(!outsideTrapezoid(iv))
      {
         newCenters.push_back(*b);
      }
      ++b;
   }
   GS.center = newCenters;
}

// ------------------------------------------------------------------------- //

void printVector(Vector<interval,2> v)
{
   rootFrame.precision(20);
   rootFrame << "[" << v[0].leftBound() << "," << Tab(30) << v[0].rightBound() << "]" << Tab(60);
   rootFrame << "[" << v[1].leftBound() << "," << Tab(90) << v[1].rightBound() << "]\n";
}

// ------------------------------------------------------------------------- //

void coordSystem(Frame& f, double sx, double sy, double x,double y)
{
   f.Clear();
   f.setWorldCoord(-sx+x,-sy+y,sx+x,sy+y);
   f.line(-sx+x,y,sx+x,y);
   f.line(x,-sy+y,x,sy+y);   
   f.Bound();
}

// ------------------------------------------------------------------------- //

void coordSystem3D(Frame *f, double s)
{
   rootFrame.Clear();
   f[0]=Frame(0,0,300,300);
   f[1]=Frame(310,0,620,300);
   f[2]=Frame(630,0,940,300);
   
   for(int i=0;i<3;i++)
   {
      coordSystem(f[i],s,s,0.,0.);
      f[i].box(-1.,-1.,1.,1.);
      f[i].Bound(GREEN);
   }
}

// ------------------------------------------------------------------------- //

void drawTrapezoid(Frame& f, int color)
{
   const static Vector<double,2> L = H0.center - H0.unstable + 2*H0.stable/10;
   const static Vector<double,2> R = H0.center + H0.stable   - 2*H0.unstable/10;
   f.jump(H0.A[0],H0.A[1]);
   f.draw(L[0],L[1],color);
   f.draw(R[0],R[1],color);
   f.draw(H0.C[0],H0.C[1],color);
   f.draw(H0.A[0],H0.A[1],color);   
}

// ------------------------------------------------------------------------- //

void drawVector(Frame &f, const IVector3D& T, int color,int d1,int d2)
{
   f.dot(T[d1].leftBound(),T[d2].leftBound());
   f.boxFill(T[d1].leftBound(),T[d2].leftBound(),T[d1].rightBound(),T[d2].rightBound(),color);
   f.box(T[d1].leftBound(),T[d2].leftBound(),T[d1].rightBound(),T[d2].rightBound(),BLACK);
}
