/// @addtogroup dynset
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file rect2setintf.h
///
/// @author The CAPD Group
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _Rect2SetIntf_H
#define _Rect2SetIntf_H

#include "dynset/c0set.h"
// set is represented as: x + C*r0 + B*r;
// move - Lohner last method - internal representation
//       C*r0 - basic 'Lipschitz part'
//       B*r - QR-decomposition of the remaining errors

namespace capd{
namespace dynset{

template<typename MatrixT>
class Rect2Set : public C0set<MatrixT>{

public:
   typedef MatrixT MatrixType;
   typedef typename MatrixType::RowVectorType VectorType;
   typedef typename MatrixType::ScalarType ScalarType;
   typedef Norm<VectorType,MatrixType> NormType;

   inline VectorType get_x () {return m_x;}
   inline VectorType get_r () {return m_r;}
   inline VectorType get_r0 () {return m_r0;}
   inline MatrixType get_B () {return m_B;}
   inline MatrixType get_C () {return m_C;}

   explicit Rect2Set(int);
   explicit Rect2Set(const VectorType& the_x);
   Rect2Set(const VectorType& the_x, const VectorType& the_r0);
   Rect2Set(const VectorType& the_x, const MatrixType& the_C, const VectorType& the_r0);
   Rect2Set(const VectorType& the_x, const MatrixType& the_C,
            const VectorType& the_r0,const MatrixType& the_B,
            const VectorType& the_r
   );
   C0set<MatrixType>* clone(void);
   C0set<MatrixType>* fresh(void);
   C0set<MatrixType>* cover(const VectorType& v);
   ScalarType size(void);
   VectorType center(void);
   void move(capd::dynsys::DynSys<MatrixType>& dynsys);
   void move(capd::dynsys::DynSys<MatrixType>& dynsys, Rect2Set& result) const;
   operator VectorType() const;
   std::string show() const;
   virtual VectorType affineTransformation(const MatrixType&, const VectorType&) const;

protected:
   VectorType m_x, m_r, m_r0;
   MatrixType m_B, m_C;
};

// inline definitions

template<typename MatrixType>
inline Rect2Set<MatrixType>::Rect2Set(int dim)
   :  m_x(dim),
      m_r(dim),
      m_r0(dim),
      m_B(MatrixType::Identity(dim)),
      m_C(MatrixType::Identity(dim))
{
   m_x.clear();
   m_r.clear();
   m_r0.clear();
}

template<typename MatrixType>
inline typename Rect2Set<MatrixType>::VectorType Rect2Set<MatrixType>::center(void)
{
   return m_x;
}

template<typename MatrixT>
inline Rect2Set<MatrixT>::operator typename Rect2Set<MatrixT>::VectorType(void) const
{
   return m_x + m_C*m_r0 + m_B*m_r;
}

template<typename MatrixType>
inline Rect2Set<MatrixType>::Rect2Set(const VectorType& the_x)
   :  m_x(the_x),
      m_r(the_x.dimension()),
      m_r0(the_x.dimension()),
      m_B(MatrixType::Identity(the_x.dimension())),
      m_C(MatrixType::Identity(the_x.dimension()))
{
   m_r.clear();
   m_r0.clear();
}

template<typename MatrixType>
inline Rect2Set<MatrixType>::Rect2Set(const VectorType& the_x,const VectorType& the_r0)
   :  m_x(the_x),
      m_r(the_x.dimension()),
      m_r0(the_r0),
      m_B(MatrixType::Identity(the_x.dimension())),
      m_C(MatrixType::Identity(the_x.dimension()))
{
   m_r.clear();
}

template<typename MatrixType>
inline Rect2Set<MatrixType>::Rect2Set(
      const VectorType& the_x,
      const MatrixType& the_C,
      const VectorType& the_r0
   ): m_x(the_x),
      m_r(the_x.dimension()),
      m_r0(the_r0),
      m_B(MatrixType::Identity(the_x.dimension())),
      m_C(the_C)
{
   m_r.clear();
}

template<typename MatrixType>
inline Rect2Set<MatrixType>::Rect2Set(
      const VectorType& the_x,
      const MatrixType &the_C,
      const VectorType& the_r0,
      const MatrixType &the_B,
      const VectorType& the_r
   ): m_x(the_x),
      m_r(the_r),
      m_r0(the_r0),
      m_B(the_B),
      m_C(the_C)
{}

template<typename MatrixType>
inline C0set<MatrixType>* Rect2Set<MatrixType>::clone(void)
{
   return new Rect2Set(*this);
}

template<typename MatrixType>
inline C0set<MatrixType>* Rect2Set<MatrixType>::fresh(void)
{
   return new Rect2Set<MatrixType>(m_x.dimension());
}

template<typename MatrixType>
inline C0set<MatrixType>* Rect2Set<MatrixType>::cover(const VectorType& v)
{
   return new Rect2Set(midVector(v),v-midVector(v));
}

}} // namespace capd::dynset

#endif // _Rect2SetIntf_H

/// @}
