/// @addtogroup dynsys
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file basicC2Taylor.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _BASICC2TAYLOR_HPP
#define _BASICC2TAYLOR_HPP

#include <string>
#include <stdexcept>
#include "dynsys/basicC2TaylorIntf.h"

namespace capd{
namespace dynsys{

//###########################################################//

template<typename MapType>
BasicC2Taylor<MapType>::BasicC2Taylor(
      const MapType& vectorField,
      int Order,
      const ScalarType& Step
   ) : BasicTaylor<MapType>(vectorField,Order,Step)
{
   m_c2MatrixCoeff = new (m_dimension) C2CoeffType[m_order+2];
}

//###########################################################//


template<typename MapType>
BasicC2Taylor<MapType>::BasicC2Taylor(
      const std::string& vectorField,
      int Order,
      const ScalarType& Step
   ) : BasicTaylor<MapType>(vectorField,Order,Step)
{
   m_c2MatrixCoeff = new (m_dimension) C2CoeffType[m_order+2];
}


//###########################################################//


template<typename MapType>
BasicC2Taylor<MapType>::BasicC2Taylor(
      const char* vectorField,
      int Order,
      const ScalarType& Step
   ) : BasicTaylor<MapType>(vectorField,Order,Step)
{
   m_c2MatrixCoeff = new (m_dimension) C2CoeffType[m_order+2];
}

//###########################################################//


template<typename MapType>
BasicC2Taylor<MapType>::BasicC2Taylor(
     const BasicC2Taylor & taylor
   ) : BasicTaylor<MapType>(taylor)
{
   m_c2MatrixCoeff = new (m_dimension) C2CoeffType[m_order+2];

   for(int i=0; i<m_order+2; ++i)
      m_c2MatrixCoeff[i] = taylor.m_c2MatrixCoeff[i];
}

//###########################################################//

template<typename MapType>
BasicC2Taylor<MapType>& BasicC2Taylor<MapType>::operator=(const BasicC2Taylor& a_t)
{
   BasicTaylor<MapType>::operator=(a_t);
   delete [] m_c2MatrixCoeff;
   m_c2MatrixCoeff = new (m_dimension) C2CoeffType[m_order+2];
   for(int i=0; i<m_order+2; ++i)
      m_c2MatrixCoeff[i] = a_t.m_c2MatrixCoeff[i];
   return *this;
}

//###########################################################//


template<typename MapType>
void BasicC2Taylor<MapType>::setOrder(int newOrder)
{
   BasicTaylor<MapType>::setOrder(newOrder);
   delete [] m_c2MatrixCoeff;
   m_c2MatrixCoeff = new (m_dimension) C2CoeffType[m_order+2];
}

//###########################################################//


template<typename MapType>
void BasicC2Taylor<MapType>::setField(const MapType& newVectorField)
{
   m_vField = newVectorField;
   m_vField.setOrder(m_order+1);
   if(m_dimension!=m_vField.getDimension())
   {
      delete [] m_coeff;
      delete [] m_matrixCoeff;
      delete [] m_c2MatrixCoeff;
      m_dimension = m_vField.getDimension();
      m_coeff = new (m_dimension) VectorType[m_order+2];
      m_matrixCoeff = new (m_dimension,m_dimension) MatrixType[m_order+2];
      m_c2MatrixCoeff = new (m_dimension) C2CoeffType[m_order+2];
   }
}

//###########################################################//

template<typename MapType>
typename BasicC2Taylor<MapType>::VectorType BasicC2Taylor<MapType>::operator()(
      const VectorType& x,
      MatrixType& jac,
      C2CoeffType& hessian
   )
{
   C2CoeffType H(m_dimension);
   VectorType result(m_dimension);
   H.clear();
   hessian.clear();
   jac.clear();
   result.clear();

   computeC2Coeff(x, MatrixType::Identity(m_dimension), H, m_order-1);

   ScalarType stepPower = 1.;
   for(int p=0;p<=m_order;++p)
   {
      typename C2CoeffType::iterator i1 = hessian.begin(),
                                     i2 = hessian.end(),
                                     j1 = m_c2MatrixCoeff[p].begin();
      while(i1!=i2)
      {
         *i1 += *j1 * stepPower;
         ++i1;
         ++j1;
      }
      jac += m_matrixCoeff[p] * stepPower;
      result += m_coeff[p] * stepPower;
      stepPower *= m_step;
   }
   return result;
}


// ####################################################################


template<typename MapType>
void BasicC2Taylor<MapType>::computeC2Coeff(
      const VectorType& x,
      const MatrixType& V,
      const C2CoeffType& H, int ile
   ) const
{
   if(ile<0)
   {
      throw std::out_of_range("C2Taylor error: negative index of Taylor coefficient in C2Taylor::computeC2Coeff");
   }

   int p,i,j,k,s,r,w,c;

// first we compute ceofficients for x
   m_coeff[0]=x;
   for(r=0;r<=ile;++r)
      m_coeff[r+1] = m_vField(m_coeff[r],r)/(r+1);


// now we compute the coefficients for the Jacobian matrix
// the coefficients will be stored in m_F
   m_vField.variational(m_coeff,m_F,ile+1);
   computeMatrixCoeff(ile,V);

   m_c2MatrixCoeff[0]=H;
   for(i=1;i<=ile+1;++i)
      m_c2MatrixCoeff[i].clear();
   C2CoeffType* temp = m_vField.computeSecondDerivatives(m_coeff,ile);

// we compute coefficients of H_{ijc} up to order 'ile+1'

   for(p=0;p<=ile;++p)
   {
      for(i=0;i<m_dimension;++i)
      {
         for(j=0;j<m_dimension;++j)
         {
            for(c=0;c<=j;++c)
            {
// -------------
               for(k=0;k<m_dimension;++k)
               {
                  for(r=0;r<m_dimension;++r)
                  {
                     for(w=0;w<=p;++w)
                     {
                        ScalarType S1(0.);
                        for(s=0;s<=w;++s)
                           S1 += temp[s](i,k,r) * m_matrixCoeff[w-s][r][c];

                        m_c2MatrixCoeff[p+1](i,j,c) += S1 * m_matrixCoeff[p-w][k][j];
                     }  // end of w-for
                  }  // end of r-for
               }  // end of k-for

               for(k=0;k<m_dimension;++k)
                  for(s=0;s<=p;++s)
                     m_c2MatrixCoeff[p+1](i,j,c) +=  m_F[s][i][k] * m_c2MatrixCoeff[p-s](k,j,c);

               m_c2MatrixCoeff[p+1](i,j,c) /= (p+1);
// -------------
            } // end of c-for
         } // end of j-for
      } //end of i-for
   }

   delete []temp;
}

}} // namespace capd::dynsys

#endif // #define _BASICC2TAYLOR_HPP
/// @}
