/// @addtogroup dynsys
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file c2taylor.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _C2TAYLOR_HPP
#define _C2TAYLOR_HPP

#include <string>
#include <stdexcept>
#include "dynsys/c2taylintf.h"

namespace capd{
namespace dynsys{

//###########################################################//

template<typename MapType>
C2Taylor<MapType>::C2Taylor(const MapType& vectorField,int Order, const ScalarType& Step)
   : BasicTaylor<MapType>(vectorField,Order,Step),
     Taylor<MapType>(vectorField,Order,Step),
     BasicC2Taylor<MapType>(vectorField,Order,Step)
{}

//###########################################################//


template<typename MapType>
C2Taylor<MapType>::C2Taylor(const std::string& vectorField, int Order, const ScalarType& Step)
   : BasicTaylor<MapType>(vectorField,Order,Step),
     Taylor<MapType>(vectorField,Order,Step),
     BasicC2Taylor<MapType>(vectorField,Order,Step)
{}


//###########################################################//


template<typename MapType>
C2Taylor<MapType>::C2Taylor(const char* vectorField, int Order, const ScalarType& Step)
   : BasicTaylor<MapType>(vectorField,Order,Step),
     Taylor<MapType>(vectorField,Order,Step),
     BasicC2Taylor<MapType>(vectorField,Order,Step)
{}

//###########################################################//

template<typename MapType>
void C2Taylor<MapType>::c2Enclosure(
         const VectorType &W1, // rough enclosure for x
         const MatrixType &W3, // rough enclosure for V
         NormType &norm,       // a logarithmic norm
         C2CoeffType& result
      ) const
{
   int i,j,c,s,r;
   MatrixType DF = m_vField[W1];
   EuclNorm<VectorType,MatrixType> tempNorm;
   ScalarType l = norm(DF).rightBound();    // l>= max x\in\W_1 \mu(\frac{df}{dx})

   C2CoeffType temp(m_dimension);
   m_vField.computeSecondDerivatives(W1,temp);

   for(j=0;j<m_dimension;++j)
   {
      for(c=j;c<m_dimension;++c)
      {
         VectorType Bjc(m_dimension);
         Bjc.clear();
         for(i=0;i<m_dimension;++i)
         {
            for(s=0;s<m_dimension;++s)
               for(r=0;r<m_dimension;++r)
                  Bjc[i] += temp(i,s,r) * W3[r][c] * W3[s][j];
         }
         ScalarType delta = tempNorm(Bjc).rightBound();
         double size;
         if(l.contains(0.0))
            size = abs(delta*abs(m_step).rightBound()).rightBound();
         else
            size = (abs(delta * (exp(l*m_step)-ScalarType(1.))/l)).rightBound();
         for(i=0;i<m_dimension;++i)
            result(i,j,c) = ScalarType(-size,size);
      } // c - loop
   } // j - loop
}

// ####################################################################


template<typename MapType>
void C2Taylor<MapType>::c2Phi(
      const VectorType& x,
      MatrixType& jac,
      C2CoeffType& hessian
   )const
{
   C2CoeffType H(m_dimension);
   H.clear();
   hessian.clear();
   jac.clear();

   computeC2Coeff(x, MatrixType::Identity(m_dimension), H, m_order-1);

   ScalarType stepPower = 1.;
   for(int p=0;p<=m_order;++p)
   {
      typename C2CoeffType::iterator i1 = hessian.begin(), i2 = hessian.end(),
           j1 = m_c2MatrixCoeff[p].begin();
      while(i1!=i2)
      {
         *i1 += *j1 * stepPower;
         ++i1;
         ++j1;
      }
      jac += m_matrixCoeff[p] * stepPower;
      stepPower *= m_step;
   }
}

// ####################################################################

template<typename MapType>
void C2Taylor<MapType>::c2Remainder(
      const VectorType& Enc,
      const MatrixType& jacEnc,
      const C2CoeffType& hessEnc,
      VectorType& Rem,
      MatrixType& jacRem,
      C2CoeffType& hessRem
   ) const
{
   computeC2Coeff(Enc, jacEnc, hessEnc, m_order);
   ScalarType fac = power(m_step,m_order+1);
   Rem = m_coeff[m_order+1] * fac;
   jacRem = m_matrixCoeff[m_order+1] * fac;

   typename C2CoeffType::iterator i1=hessRem.begin(), i2=hessRem.end(), j=m_c2MatrixCoeff[m_order+1].begin();
   while(i1!=i2)
   {
      *i1 = *j * fac;
      ++i1;
      ++j;
   }
}

}} // namespace capd::dynsys

#endif // #define _C2TAYLOR_HPP
/// @}
