/// @addtogroup dynsys
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file enclosure.h
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

/** Author: Daniel Wilczak, 2001-2004 */

#ifndef _ENCLOSURE_H_
#define _ENCLOSURE_H_

#include <sstream>
#include <string>
#include <stdexcept>

#include "dynsys/taylorex.h"

namespace capd{
namespace dynsys{


// the function finds an enclosure for \varphi([0,step],x)
template<typename MapType>
typename MapType::VectorType enclosure(  MapType const & vField, 
                                         typename MapType::MatrixType::RowVectorType const & x, 
                                         typename MapType::ScalarType const & step
                                      ) {
  typedef typename MapType::ScalarType ScalarType;
  typedef typename MapType::VectorType VectorType;

  ScalarType trialStep = ScalarType(-0.2,1.2)*step;
  int dimension = x.dimension();
  VectorType y(dimension),z(dimension),Small(dimension);
  
  ScalarType h = ScalarType(0,1) * step;
  
  bool found = false;
  int counter=0,
      limit=10,    // maximum numbers of tries to find enclosure
      i;
  typename ScalarType::BoundType multf = 1.5;   // factor to multiply coordinates if inclusion fails
  
  for(i=0;i<dimension;++i)
    Small[i] = ScalarType(-1,1)*1e-15;
  
  VectorType val = vField(x);
  z = x + trialStep * val;
  z += Small; // to be sure that z has noempty interior
  
  while((!found) && (counter<limit)){
    counter++;
    y = x + h * vField(z);
    found = true;
    for(i=0;i< dimension;++i){
      if(!(y[i].subsetInterior(z[i]))){
        found = false;
        z[i] = y[i];
        ScalarType s;
        z[i].split(s);
        s = multf*s;
        z[i] += s;
      } 
    }
  }
  
  if(!found)
    throw TaylorException<VectorType>("Error: cannot find enclosure guaranteeing bounds",x,step);

  return y;
}


//###########################################################//

}} //namespace capd::dynsys

#endif // #define _ENCLOSURE_H_

/// @}
