/// @addtogroup map
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file CnCoeff.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

/** Author: Daniel Wilczak,  2006 */

#ifndef _CNCOEFF_HPP
#define _CNCOEFF_HPP
#include <algorithm>

namespace capd{
namespace map{

// -----------------------------------------------------------------

template<typename MatrixT>
CnCoeff<MatrixT>::operator typename CnCoeff<MatrixT>::VectorType() const
{
   VectorType result(m_dim);
   typename VectorType::iterator b=result.begin(), e=result.end();
   const_iterator i=begin();
   int stride = newton(m_dim+m_rank,m_dim);
   while(b!=e)
   {
      *b = *i;
      ++b;
      i+=stride;
   }
   return result;
}

// -----------------------------------------------------------------

template<typename MatrixT>
CnCoeff<MatrixT>::operator typename CnCoeff<MatrixT>::MatrixType() const
{
   MatrixType result(m_dim,m_dim);
   typename MatrixType::ContainerIterator b=result.cbegin(), e=result.cend();
   int i=0;
   while(b!=e)
   {
      const_iterator b1=begin(i,1), e1=end(i,1);
      while(b1!=e1)
      {
         *b = *b1;
         ++b;
         ++b1;
      }
      ++i;
   }
   return result;
}

// -----------------------------------------------------------------
template<typename MatrixT>
void CnCoeff<MatrixT>::setMatrix(const MatrixType& m)
{
   typename MatrixType::const_ContainerIterator b=m.cbegin(), e=m.cend();
   int i=0;
   while(b!=e)
   {
      iterator b1=begin(i,1), e1=end(i,1);
      while(b1!=e1)
      {
         *b1 = *b;
         ++b;
         ++b1;
      }
      ++i;
   }
}


// -----------------------------------------------------------------

template<typename MatrixType>
void CnCoeff<MatrixType>::clear()
{
   iterator b=begin(), e=end();
   while(b!=e)
   {
      *b = static_cast<ScalarType>(0);
      ++b;
   }
}

// -----------------------------------------------------------------

template<typename MatrixType>
typename CnCoeff<MatrixType>::VectorType CnCoeff<MatrixType>::operator()(const VectorType& v) const
{
   VectorType result = VectorType(*this);
   
   for(int i=1;i<=m_rank;++i)
   {
      Multipointer mp = this->first(i);
      do{
         typename VectorType::ScalarType temp = power(v,mp);
         long f = mp.factorial();
         result += temp* (*this)(mp)/f;
      }while(next(mp));
   }
   return result;
}

}} // the end of the namespace capd::map

// -----------------------------------------------------------------

template<typename MatrixType>
capd::map::CnCoeff<MatrixType> operator+(const capd::map::CnCoeff<MatrixType>& x, const capd::map::CnCoeff<MatrixType>& y)
{
   if(x.dimension()!=y.dimension())
      throw std::runtime_error("CnCoeff operator+: incompatible dimensions of power series");
   capd::map::CnCoeff<MatrixType> result(x);
   typename capd::map::CnCoeff<MatrixType>::iterator i = result.begin();
   typename capd::map::CnCoeff<MatrixType>::const_iterator b = y.begin(), e=y.end();
   while(b!=e)
   {
      (*i) += (*b);
      ++b;
      ++i;
   }
   return result;
}

// -----------------------------------------------------------------

template<typename MatrixType>
capd::map::CnCoeff<MatrixType> operator-(const capd::map::CnCoeff<MatrixType>& x, const capd::map::CnCoeff<MatrixType>& y)
{
   if(x.dimension()!=y.dimension())
      throw std::runtime_error("CnCoeff operator-: incompatible dimensions of power series");
   capd::map::CnCoeff<MatrixType> result(x);
   typename capd::map::CnCoeff<MatrixType>::iterator i = result.begin();
   typename capd::map::CnCoeff<MatrixType>::const_iterator b = y.begin(), e=y.end();
   while(b!=e)
   {
      (*i) -= (*b);
      ++b;
      ++i;
   }
   return result;
}

// -----------------------------------------------------------------

template<typename MatrixType>
typename MatrixType::ScalarType computeProduct(
      const capd::map::CnCoeff<MatrixType>& second,
      const capd::vectalg::Multiindex& mi,
      const capd::vectalg::Multipointer& a,
      int p, int k
   )
{
   typedef typename MatrixType::ScalarType Scalar;
   typedef capd::vectalg::Multiindex Multiindex;
   typedef capd::vectalg::Multipointer Multipointer;

   Scalar result(0);
   const typename Multipointer::IndicesSet& is = Multipointer::generateList(p,k);
   typename Multipointer::IndicesSet::const_iterator b=is.begin(), e=is.end();

   while(b!=e)
   {
      typename Multipointer::MultipointersVector::const_iterator bt = b->begin(), et=b->end();
      typename Multiindex::const_iterator ib= mi.begin();

      Multipointer delta = a.subMultipointer(*bt);
      Scalar temp = second(*ib,delta);
      ++bt;
      ++ib;
      while(bt!=et)
      {
         Multipointer delta = a.subMultipointer(*bt);
         temp *= second(*ib,delta);
         ++bt;
         ++ib;
      }
      ++b;
      result += temp;
   }
   return result;
}

// -----------------------------------------------------------------

template<typename MatrixType>
void computeComposition(
      const capd::map::CnCoeff<MatrixType>& first,
      const capd::map::CnCoeff<MatrixType>& second,
      capd::map::CnCoeff<MatrixType>& result,
      const capd::vectalg::Multipointer& a,
      bool nonlinearOnly
   )
{
   typedef typename MatrixType::ScalarType Scalar;
   typedef capd::vectalg::Multiindex Multiindex;
   typedef capd::vectalg::Multipointer Multipointer;
   typename Multiindex::IndicesSet listIndices;   

   Multiindex::generateList(result.dimension(),result.rank(),listIndices);
   int p = a.module();

   result(a).clear();
   int mink = nonlinearOnly ? 2 : 1;
   for(int k=mink;k<=p;++k)
   {
      typename Multiindex::MultiindexVector::iterator b = listIndices[k-1].begin(), e=listIndices[k-1].end();
      while(b!=e)
      {
         Multipointer mp(b->dimension(),b->begin());
         std::sort(mp.begin(),mp.end());
         Scalar product = computeProduct(second,*b,a,p,k);
         result(a) += first(mp) * product;
         ++b;
      }
   }
}

// -----------------------------------------------------------------


template<typename MatrixType>
void substitutionPowerSeries(
      const capd::map::CnCoeff<MatrixType>& first,
      const capd::map::CnCoeff<MatrixType>& second,
      capd::map::CnCoeff<MatrixType>& result, bool nonlinearOnly
   )
// the function computes substitution of two power series truncated to the same order as 'first' and 'second'
// as a result we obtain an expansion of first(second)
{
   typedef typename MatrixType::ScalarType Scalar;
   typedef capd::vectalg::Multiindex Multiindex;
   typedef capd::vectalg::Multipointer Multipointer;

   for(int i=1;i<=first.rank();++i)
   {
      Multipointer a = first.first(i);
      do{
         computeComposition(first,second,result,a,nonlinearOnly);
      }while(first.next(a));
   }
   
   if(!nonlinearOnly)
      result() = first();
}

#endif

/// @}
